//! Track-based composition module
//!
//! This module provides direct control over musical tracks and patterns

use crate::config::{CompositionConfig, TrackConfig, TrackPattern};
use crate::core::{
    Composition as CoreComposition, CompositionParams, CompositionStyle, InstrumentType, Note,
    Track,
};
use crate::scales::{Scale, ScaleType};

/// Track-based composition
#[derive(Debug, Clone)]
pub struct Composition {
    pub key: u8,
    pub scale_type: ScaleType,
    pub tempo: f32,
    pub time_signature: (u8, u8),
    pub measures: usize,
    pub track_configs: Vec<TrackConfig>,
    pub tracks: Vec<Track>,
    pub total_duration: f32,
}

impl Composition {
    /// Create a new track composition from configuration
    pub fn from_config(config: CompositionConfig) -> Result<Self, String> {
        let key = config.composition.key.to_midi()?;
        let scale_type = parse_scale_type(&config.composition.scale)?;

        let total_duration =
            config.composition.measures as f32 * config.composition.time_signature.numerator as f32;

        Ok(Self {
            key,
            scale_type,
            tempo: config.composition.tempo,
            time_signature: (
                config.composition.time_signature.numerator,
                config.composition.time_signature.denominator,
            ),
            measures: config.composition.measures,
            track_configs: config.tracks,
            tracks: Vec::new(),
            total_duration,
        })
    }

    /// Generate the composition by processing all tracks
    pub fn generate(&mut self) -> Result<(), String> {
        self.tracks.clear();

        let scale = Scale::new(self.scale_type, self.key);

        for track_config in &self.track_configs {
            let track = self.generate_track(track_config, &scale)?;
            self.tracks.push(track);
        }

        Ok(())
    }

    /// Generate a track from a track configuration
    fn generate_track(&self, track: &TrackConfig, scale: &Scale) -> Result<Track, String> {
        let notes = match track.pattern.pattern_type.as_str() {
            "custom" => self.generate_custom_pattern(&track.pattern)?,
            "chord" => self.generate_chord_pattern(&track.pattern, scale)?,
            "arpeggio" => self.generate_arpeggio_pattern(&track.pattern, scale)?,
            "sequence" => self.generate_sequence_pattern(&track.pattern, scale)?,
            _ => {
                return Err(format!(
                    "Unknown pattern type: {}",
                    track.pattern.pattern_type
                ));
            }
        };

        let instrument_type = match track.name.to_lowercase().as_str() {
            name if name.contains("bass") => InstrumentType::Bass,
            name if name.contains("lead") => InstrumentType::Lead,
            name if name.contains("pad") => InstrumentType::Pad,
            name if name.contains("arp") => InstrumentType::Arp,
            name if name.contains("drum")
                || name.contains("kick")
                || name.contains("snare")
                || name.contains("hat") =>
            {
                InstrumentType::Percussion
            }
            _ => InstrumentType::Lead,
        };

        Ok(Track {
            name: track.name.clone(),
            octave: 4, // Will be overridden by note specifications
            notes,
            volume: track.volume,
            instrument_type,
        })
    }

    /// Generate notes from a custom pattern
    fn generate_custom_pattern(&self, pattern: &TrackPattern) -> Result<Vec<Note>, String> {
        let mut notes = Vec::new();
        let loop_count = (self.total_duration / pattern.loop_length).ceil() as usize;

        for loop_iteration in 0..loop_count {
            let loop_offset = loop_iteration as f32 * pattern.loop_length;

            for step in &pattern.steps {
                let absolute_time = loop_offset + step.time;

                // Don't generate notes beyond the composition length
                if absolute_time >= self.total_duration {
                    break;
                }

                let midi_note = parse_note(&step.note)?;

                notes.push(Note::new(
                    midi_note,
                    absolute_time,
                    step.duration,
                    step.velocity,
                ));
            }
        }

        Ok(notes)
    }

    /// Generate notes from a chord pattern
    fn generate_chord_pattern(
        &self,
        pattern: &TrackPattern,
        _scale: &Scale,
    ) -> Result<Vec<Note>, String> {
        let mut notes = Vec::new();
        let loop_count = (self.total_duration / pattern.loop_length).ceil() as usize;

        for loop_iteration in 0..loop_count {
            let loop_offset = loop_iteration as f32 * pattern.loop_length;

            for chord_step in &pattern.chord_progression {
                let absolute_time = loop_offset + chord_step.time;

                if absolute_time >= self.total_duration {
                    break;
                }

                let chord_notes = parse_chord(&chord_step.chord, pattern.octave.unwrap_or(4))?;

                for &chord_note in &chord_notes {
                    notes.push(Note::new(
                        chord_note,
                        absolute_time,
                        chord_step.duration,
                        0.7, // Default chord velocity
                    ));
                }
            }
        }

        Ok(notes)
    }

    /// Generate notes from an arpeggio pattern
    fn generate_arpeggio_pattern(
        &self,
        pattern: &TrackPattern,
        _scale: &Scale,
    ) -> Result<Vec<Note>, String> {
        let mut notes = Vec::new();
        let loop_count = (self.total_duration / pattern.loop_length).ceil() as usize;

        for loop_iteration in 0..loop_count {
            let loop_offset = loop_iteration as f32 * pattern.loop_length;

            for chord_step in &pattern.chord_progression {
                let chord_start_time = loop_offset + chord_step.time;

                if chord_start_time >= self.total_duration {
                    break;
                }

                let chord_notes = parse_chord(&chord_step.chord, pattern.octave.unwrap_or(4))?;
                let note_duration = chord_step.duration / chord_notes.len() as f32;

                for (i, &chord_note) in chord_notes.iter().enumerate() {
                    let note_time = chord_start_time + (i as f32 * note_duration);

                    if note_time >= self.total_duration {
                        break;
                    }

                    notes.push(Note::new(
                        chord_note,
                        note_time,
                        note_duration * 0.9, // Slightly staccato
                        0.6,
                    ));
                }
            }
        }

        Ok(notes)
    }

    /// Generate notes from a sequence pattern (placeholder)
    fn generate_sequence_pattern(
        &self,
        pattern: &TrackPattern,
        _scale: &Scale,
    ) -> Result<Vec<Note>, String> {
        // For now, treat sequence patterns like custom patterns
        // This can be extended to support step sequencer-style patterns
        self.generate_custom_pattern(pattern)
    }

    /// Get all notes from all tracks sorted by start time
    pub fn get_all_notes(&self) -> Vec<Note> {
        let mut all_notes = Vec::new();

        for track in &self.tracks {
            all_notes.extend(track.notes.clone());
        }

        all_notes.sort_by(|a, b| a.start_time.partial_cmp(&b.start_time).unwrap());
        all_notes
    }

    /// Get composition statistics
    pub fn get_stats(&self) -> CompositionStats {
        let total_notes = self.tracks.iter().map(|t| t.notes.len()).sum();

        CompositionStats {
            total_notes,
            total_duration: self.total_duration,
            track_count: self.tracks.len(),
            measures: self.measures,
            tempo: self.tempo,
        }
    }

    /// Convert to old Composition format for audio export compatibility
    pub fn to_composition(&self) -> CoreComposition {
        let params = CompositionParams {
            style: CompositionStyle::Electronic, // Default fallback
            key: self.key,
            scale_type: self.scale_type,
            tempo: self.tempo,
            time_signature: self.time_signature,
            measures: self.measures,
            track_count: self.tracks.len().max(1),
            complexity: 0.6, // Default values since tracks don't use these
            harmony_density: 0.7,
            rhythmic_density: 0.7,
        };

        let mut composition = CoreComposition::new(params);
        composition.tracks = self.tracks.clone();
        composition.total_duration = self.total_duration;

        composition
    }
}

/// Statistics about a track composition
#[derive(Debug, Clone)]
pub struct CompositionStats {
    pub total_notes: usize,
    pub total_duration: f32,
    pub track_count: usize,
    pub measures: usize,
    pub tempo: f32,
}

/// Parse a note string into a MIDI note number
fn parse_note(note_str: &str) -> Result<u8, String> {
    // Try parsing as MIDI number first
    if let Ok(midi_num) = note_str.parse::<u8>() {
        if midi_num <= 127 {
            return Ok(midi_num);
        } else {
            return Err(format!(
                "MIDI note number {} out of range (0-127)",
                midi_num
            ));
        }
    }

    // Parse as note name (e.g., "C4", "F#3", "Bb2")
    parse_note_name(note_str)
}

/// Parse note name to MIDI number
fn parse_note_name(name: &str) -> Result<u8, String> {
    let name = name.trim().to_uppercase();

    // Extract note, accidental, and octave
    let (note_part, octave_str) = name.split_at(
        name.find(|c: char| c.is_ascii_digit())
            .ok_or_else(|| format!("Invalid note name: {} (no digit found)", name))?,
    );

    let octave: i32 = octave_str
        .parse()
        .map_err(|_| format!("Invalid octave: {}", octave_str))?;

    // Base MIDI numbers for C in each octave
    let base_midi = (octave + 1) * 12;

    // Parse note and accidental
    let (note_char, accidental) = if note_part.ends_with('#') || note_part.ends_with('♯') {
        (&note_part[0..1], 1)
    } else if note_part.ends_with('b') || note_part.ends_with('♭') || note_part.ends_with('B') {
        (&note_part[0..1], -1)
    } else {
        (note_part, 0)
    };

    // Note offsets from C
    let note_offset = match note_char {
        "C" => 0,
        "D" => 2,
        "E" => 4,
        "F" => 5,
        "G" => 7,
        "A" => 9,
        "B" => 11,
        _ => return Err(format!("Invalid note: {}", note_char)),
    };

    let midi = base_midi + note_offset as i32 + accidental as i32;
    if midi < 0 || midi > 127 {
        return Err(format!("MIDI note {} out of range (0-127)", midi));
    }

    Ok(midi as u8)
}

/// Parse a chord name into MIDI note numbers
fn parse_chord(chord_name: &str, octave: u8) -> Result<Vec<u8>, String> {
    // Simple chord parsing - this can be extended for more complex chords
    let chord_name = chord_name.trim();

    // Extract root note
    let root_char = chord_name
        .chars()
        .next()
        .ok_or_else(|| format!("Empty chord name: {}", chord_name))?;

    let mut pos = 1;
    let mut root_accidental = 0;

    // Check for accidental
    if let Some(second_char) = chord_name.chars().nth(1) {
        match second_char {
            '#' | '♯' => {
                root_accidental = 1;
                pos = 2;
            }
            'b' | '♭' => {
                root_accidental = -1;
                pos = 2;
            }
            _ => {}
        }
    }

    // Get root MIDI note
    let root_offset = match root_char.to_uppercase().next().unwrap() {
        'C' => 0,
        'D' => 2,
        'E' => 4,
        'F' => 5,
        'G' => 7,
        'A' => 9,
        'B' => 11,
        _ => return Err(format!("Invalid root note: {}", root_char)),
    };

    let root_midi = ((octave as i32 + 1) * 12 + root_offset + root_accidental) as u8;

    // Parse chord quality
    let quality = &chord_name[pos..];

    let intervals = match quality.to_lowercase().as_str() {
        "" | "maj" | "major" => vec![0, 4, 7],  // Major triad
        "m" | "min" | "minor" => vec![0, 3, 7], // Minor triad
        "7" | "dom7" => vec![0, 4, 7, 10],      // Dominant 7th
        "maj7" | "M7" => vec![0, 4, 7, 11],     // Major 7th
        "m7" | "min7" => vec![0, 3, 7, 10],     // Minor 7th
        "dim" | "°" => vec![0, 3, 6],           // Diminished
        "aug" | "+" => vec![0, 4, 8],           // Augmented
        "sus2" => vec![0, 2, 7],                // Suspended 2nd
        "sus4" => vec![0, 5, 7],                // Suspended 4th
        _ => vec![0, 4, 7],                     // Default to major if unknown
    };

    let mut chord_notes = Vec::new();
    for &interval in &intervals {
        let note = root_midi + interval;
        if note <= 127 {
            chord_notes.push(note);
        }
    }

    Ok(chord_notes)
}

/// Parse scale type from string
fn parse_scale_type(scale_str: &str) -> Result<ScaleType, String> {
    match scale_str.to_lowercase().as_str() {
        "major" => Ok(ScaleType::Major),
        "minor" => Ok(ScaleType::Minor),
        "dorian" => Ok(ScaleType::Dorian),
        "phrygian" => Ok(ScaleType::Phrygian),
        "lydian" => Ok(ScaleType::Lydian),
        "mixolydian" => Ok(ScaleType::Mixolydian),
        "aeolian" => Ok(ScaleType::Aeolian),
        "locrian" => Ok(ScaleType::Locrian),
        "pentatonic" => Ok(ScaleType::Pentatonic),
        "blues" => Ok(ScaleType::Blues),
        "chromatic" => Ok(ScaleType::Chromatic),
        _ => Err(format!("Unknown scale: {}", scale_str)),
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_note_parsing() {
        assert_eq!(parse_note("60").unwrap(), 60);
        assert_eq!(parse_note("C4").unwrap(), 60);
        assert_eq!(parse_note("A4").unwrap(), 69);
        assert_eq!(parse_note("C#4").unwrap(), 61);
        assert_eq!(parse_note("Bb3").unwrap(), 58);
    }

    #[test]
    fn test_chord_parsing() {
        let c_major = parse_chord("C", 4).unwrap();
        assert_eq!(c_major, vec![60, 64, 67]); // C4, E4, G4

        let d_minor = parse_chord("Dm", 4).unwrap();
        assert_eq!(d_minor, vec![62, 65, 69]); // D4, F4, A4

        let g7 = parse_chord("G7", 4).unwrap();
        assert_eq!(g7, vec![67, 71, 74, 77]); // G4, B4, D5, F5
    }

    #[test]
    fn test_scale_type_parsing() {
        assert_eq!(parse_scale_type("major").unwrap(), ScaleType::Major);
        assert_eq!(parse_scale_type("minor").unwrap(), ScaleType::Minor);
        assert_eq!(parse_scale_type("blues").unwrap(), ScaleType::Blues);
    }
}
