//! JSON configuration module for composition specifications
//!
//! This module provides structures and parsing for JSON-based composition definitions,
//! allowing users to specify complete compositions in a declarative format.

use crate::core::{CompositionParams, CompositionStyle};
use crate::scales::ScaleType;
use serde::{Deserialize, Serialize};
use std::fs;
use std::path::Path;

/// Main configuration structure for a composition
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct CompositionConfig {
    /// Metadata about the composition
    #[serde(default)]
    pub metadata: Metadata,

    /// Global composition settings
    pub composition: CompositionSettings,

    /// Track definitions
    #[serde(default)]
    pub tracks: Vec<TrackConfig>,

    /// Optional section definitions for structured compositions
    #[serde(default)]
    pub sections: Vec<SectionConfig>,

    /// Export settings
    #[serde(default)]
    pub export: ExportSettings,
}

/// Metadata about the composition
#[derive(Debug, Clone, Serialize, Deserialize, Default)]
pub struct Metadata {
    #[serde(default)]
    pub title: String,

    #[serde(default)]
    pub artist: String,

    #[serde(default)]
    pub description: String,

    #[serde(default)]
    pub tags: Vec<String>,
}

/// Core composition settings
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct CompositionSettings {
    /// Musical key (MIDI note number or note name)
    pub key: KeySpec,

    /// Scale type
    pub scale: String,

    /// Tempo in BPM
    pub tempo: f32,

    /// Time signature
    #[serde(default = "default_time_signature")]
    pub time_signature: TimeSignature,

    /// Number of measures
    pub measures: usize,

    /// Complexity (0.0 to 1.0)
    #[serde(default = "default_complexity")]
    pub complexity: f32,

    /// Harmony density (0.0 to 1.0)
    #[serde(default = "default_harmony_density")]
    pub harmony_density: f32,

    /// Rhythmic density (0.0 to 1.0)
    #[serde(default = "default_rhythmic_density")]
    pub rhythmic_density: f32,

    /// Random seed for reproducible generation
    #[serde(default)]
    pub seed: Option<u64>,
}

/// Key specification - can be either a MIDI number or note name
#[derive(Debug, Clone, Serialize, Deserialize)]
#[serde(untagged)]
pub enum KeySpec {
    Midi(u8),
    NoteName(String),
}

impl KeySpec {
    /// Convert to MIDI note number
    pub fn to_midi(&self) -> Result<u8, String> {
        match self {
            KeySpec::Midi(n) => Ok(*n),
            KeySpec::NoteName(name) => parse_note_name(name),
        }
    }
}

/// Time signature specification
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct TimeSignature {
    pub numerator: u8,
    pub denominator: u8,
}

impl Default for TimeSignature {
    fn default() -> Self {
        Self {
            numerator: 4,
            denominator: 4,
        }
    }
}

/// Track configuration for direct control over individual instruments
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct TrackConfig {
    /// Track name
    pub name: String,

    /// Instrument type
    pub instrument: String,

    /// Volume (0.0 to 1.0)
    #[serde(default = "default_volume")]
    pub volume: f32,

    /// Pattern definition
    pub pattern: TrackPattern,

    /// Effects to apply
    #[serde(default)]
    pub effects: Vec<EffectConfig>,
}

/// Pattern definition for a track
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct TrackPattern {
    /// Pattern type
    #[serde(rename = "type")]
    pub pattern_type: String,

    /// Pattern steps for custom patterns
    #[serde(default)]
    pub steps: Vec<PatternStep>,

    /// Loop length in beats
    #[serde(default = "default_loop_length")]
    pub loop_length: f32,

    /// Chord progression for chord patterns
    #[serde(default)]
    pub chord_progression: Vec<ChordStep>,

    /// Chord voicing style
    #[serde(default)]
    pub voicing: Option<String>,

    /// Octave for chord patterns
    #[serde(default)]
    pub octave: Option<u8>,
}

/// Individual pattern step
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct PatternStep {
    /// Time in beats
    pub time: f32,

    /// Note (MIDI number or note name)
    pub note: String,

    /// Duration in beats
    pub duration: f32,

    /// Velocity (0.0 to 1.0)
    pub velocity: f32,
}

/// Chord progression step
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct ChordStep {
    /// Time in beats
    pub time: f32,

    /// Chord name
    pub chord: String,

    /// Duration in beats
    pub duration: f32,
}

/// Section definition for structured compositions
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct SectionConfig {
    /// Section name (e.g., "intro", "verse", "chorus")
    pub name: String,

    /// Number of measures in this section
    pub measures: usize,

    /// Optional tempo change
    #[serde(default)]
    pub tempo: Option<f32>,

    /// Optional key change
    #[serde(default)]
    pub key: Option<KeySpec>,

    /// Optional scale change
    #[serde(default)]
    pub scale: Option<String>,

    /// Complexity override for this section
    #[serde(default)]
    pub complexity: Option<f32>,

    /// Number of times to repeat this section
    #[serde(default = "default_repeat")]
    pub repeat: usize,
}

/// Pattern configuration
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct PatternConfig {
    /// Pattern type
    #[serde(rename = "type")]
    pub pattern_type: String,

    /// Pattern-specific parameters
    #[serde(flatten)]
    pub params: serde_json::Value,
}

/// Effect configuration
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct EffectConfig {
    /// Effect type
    #[serde(rename = "type")]
    pub effect_type: String,

    /// Effect parameters
    #[serde(flatten)]
    pub params: serde_json::Value,
}

/// Export settings
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct ExportSettings {
    /// Output filename (without extension)
    #[serde(default = "default_filename")]
    pub filename: String,

    /// Export format
    #[serde(default = "default_format")]
    pub format: String,

    /// Sample rate
    #[serde(default = "default_sample_rate")]
    pub sample_rate: u32,

    /// Bit depth
    #[serde(default = "default_bit_depth")]
    pub bit_depth: u16,

    /// Stereo or mono
    #[serde(default = "default_stereo")]
    pub stereo: bool,

    /// Maximum duration in seconds (None for full composition)
    #[serde(default)]
    pub max_duration: Option<f32>,

    /// Generate variations
    #[serde(default)]
    pub variations: Option<VariationSettings>,
}

/// Variation generation settings
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct VariationSettings {
    /// Number of variations to generate
    pub count: usize,

    /// Variation parameters
    #[serde(default)]
    pub vary_complexity: bool,

    #[serde(default)]
    pub vary_rhythm: bool,

    #[serde(default)]
    pub vary_harmony: bool,

    #[serde(default)]
    pub vary_tempo: bool,
}

impl Default for ExportSettings {
    fn default() -> Self {
        Self {
            filename: default_filename(),
            format: default_format(),
            sample_rate: default_sample_rate(),
            bit_depth: default_bit_depth(),
            stereo: default_stereo(),
            max_duration: None,
            variations: None,
        }
    }
}

// Default value functions
fn default_complexity() -> f32 {
    0.6
}
fn default_harmony_density() -> f32 {
    0.7
}
fn default_rhythmic_density() -> f32 {
    0.8
}
fn default_volume() -> f32 {
    1.0
}
fn default_loop_length() -> f32 {
    4.0
}
fn default_repeat() -> usize {
    1
}
fn default_filename() -> String {
    "output".to_string()
}
fn default_format() -> String {
    "wav".to_string()
}
fn default_sample_rate() -> u32 {
    44100
}
fn default_bit_depth() -> u16 {
    16
}
fn default_stereo() -> bool {
    false
}
fn default_time_signature() -> TimeSignature {
    TimeSignature::default()
}

/// Parse note name to MIDI number
fn parse_note_name(name: &str) -> Result<u8, String> {
    let name = name.trim().to_uppercase();

    // Extract note, accidental, and octave
    let (note_part, octave_str) = name.split_at(
        name.find(|c: char| c.is_ascii_digit())
            .ok_or_else(|| format!("Invalid note name: {}", name))?,
    );

    let octave: i32 = octave_str
        .parse()
        .map_err(|_| format!("Invalid octave: {}", octave_str))?;

    // Base MIDI numbers for C in each octave
    let base_midi = (octave + 1) * 12;

    // Parse note and accidental
    let (note_char, accidental) = if note_part.ends_with('#') || note_part.ends_with('♯') {
        (&note_part[0..1], 1)
    } else if note_part.ends_with('b') || note_part.ends_with('♭') || note_part.ends_with('B') {
        (&note_part[0..1], -1)
    } else {
        (note_part, 0)
    };

    // Note offsets from C
    let note_offset = match note_char {
        "C" => 0,
        "D" => 2,
        "E" => 4,
        "F" => 5,
        "G" => 7,
        "A" => 9,
        "B" => 11,
        _ => return Err(format!("Invalid note: {}", note_char)),
    };

    let midi = base_midi + note_offset as i32 + accidental as i32;
    if midi < 0 || midi > 127 {
        return Err(format!("MIDI note {} out of range (0-127)", midi));
    }

    Ok(midi as u8)
}

impl CompositionConfig {
    /// Load configuration from a JSON file
    pub fn from_file<P: AsRef<Path>>(path: P) -> Result<Self, String> {
        let content =
            fs::read_to_string(path).map_err(|e| format!("Failed to read config file: {}", e))?;

        Self::from_json(&content)
    }

    /// Parse configuration from JSON string
    pub fn from_json(json: &str) -> Result<Self, String> {
        serde_json::from_str(json).map_err(|e| format!("Failed to parse JSON: {}", e))
    }

    /// Save configuration to a JSON file
    pub fn to_file<P: AsRef<Path>>(&self, path: P) -> Result<(), String> {
        let json = self.to_json_pretty()?;
        fs::write(path, json).map_err(|e| format!("Failed to write config file: {}", e))
    }

    /// Serialize configuration to pretty JSON
    pub fn to_json_pretty(&self) -> Result<String, String> {
        serde_json::to_string_pretty(self)
            .map_err(|e| format!("Failed to serialize to JSON: {}", e))
    }

    /// Convert to composition parameters (deprecated - use track-based generation instead)
    pub fn to_composition_params(&self) -> Result<CompositionParams, String> {
        let scale_type = match self.composition.scale.to_lowercase().as_str() {
            "major" => ScaleType::Major,
            "minor" => ScaleType::Minor,
            "dorian" => ScaleType::Dorian,
            "phrygian" => ScaleType::Phrygian,
            "lydian" => ScaleType::Lydian,
            "mixolydian" => ScaleType::Mixolydian,
            "aeolian" => ScaleType::Aeolian,
            "locrian" => ScaleType::Locrian,
            "pentatonic" => ScaleType::Pentatonic,
            "blues" => ScaleType::Blues,
            "chromatic" => ScaleType::Chromatic,
            _ => return Err(format!("Unknown scale: {}", self.composition.scale)),
        };

        let key = self.composition.key.to_midi()?;

        Ok(CompositionParams {
            style: CompositionStyle::Electronic, // Default fallback
            key,
            scale_type,
            tempo: self.composition.tempo,
            time_signature: (
                self.composition.time_signature.numerator,
                self.composition.time_signature.denominator,
            ),
            measures: self.composition.measures,
            track_count: self.tracks.len().max(1),
            complexity: self.composition.complexity,
            harmony_density: self.composition.harmony_density,
            rhythmic_density: self.composition.rhythmic_density,
        })
    }

    /// Create an example configuration
    pub fn example() -> Self {
        Self {
            metadata: Metadata {
                title: "Example Track Composition".to_string(),
                artist: "Track Composer".to_string(),
                description: "An example track-based composition".to_string(),
                tags: vec!["lofi".to_string(), "chill".to_string()],
            },
            composition: CompositionSettings {
                key: KeySpec::NoteName("C4".to_string()),
                scale: "minor".to_string(),
                tempo: 85.0,
                time_signature: TimeSignature {
                    numerator: 4,
                    denominator: 4,
                },
                measures: 16,
                complexity: 0.6,
                harmony_density: 0.7,
                rhythmic_density: 0.7,
                seed: Some(42),
            },
            tracks: vec![TrackConfig {
                name: "bass".to_string(),
                instrument: "sine".to_string(),
                volume: 0.9,
                pattern: TrackPattern {
                    pattern_type: "custom".to_string(),
                    steps: vec![
                        PatternStep {
                            time: 0.0,
                            note: "C2".to_string(),
                            duration: 0.75,
                            velocity: 0.9,
                        },
                        PatternStep {
                            time: 2.0,
                            note: "G2".to_string(),
                            duration: 0.75,
                            velocity: 0.8,
                        },
                    ],
                    loop_length: 4.0,
                    chord_progression: vec![],
                    voicing: None,
                    octave: None,
                },
                effects: vec![EffectConfig {
                    effect_type: "lowpass".to_string(),
                    params: serde_json::json!({
                        "cutoff": 400.0,
                        "resonance": 1.8
                    }),
                }],
            }],
            sections: vec![],
            export: ExportSettings {
                filename: "track_composition".to_string(),
                format: "wav".to_string(),
                sample_rate: 44100,
                bit_depth: 16,
                stereo: true,
                max_duration: None,
                variations: None,
            },
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_parse_note_names() {
        assert_eq!(parse_note_name("C4").unwrap(), 60);
        assert_eq!(parse_note_name("A4").unwrap(), 69);
        assert_eq!(parse_note_name("C#4").unwrap(), 61);
        assert_eq!(parse_note_name("Bb3").unwrap(), 58);
        assert_eq!(parse_note_name("G5").unwrap(), 79);
    }

    #[test]
    fn test_config_serialization() {
        let config = CompositionConfig::example();
        let json = config.to_json_pretty().unwrap();
        let parsed = CompositionConfig::from_json(&json).unwrap();

        assert_eq!(parsed.composition.tempo, config.composition.tempo);
    }

    #[test]
    fn test_to_composition_params() {
        let config = CompositionConfig::example();
        let params = config.to_composition_params().unwrap();

        assert_eq!(params.key, 60); // C4
        assert_eq!(params.tempo, 85.0);
        assert_eq!(params.measures, 16);
    }
}
