//! Generate electronic music without AI
//!
//! This library provides tools for generating electronic music
//! with synthesis, sequencing, and composition algorithms.

pub mod audio;
pub mod composition;
pub mod config;
pub mod core;
pub mod effects;
pub mod patterns;
pub mod scales;
pub mod sequencer;
pub mod synthesis;

use std::f32::consts::PI;

/// Sample rate for audio generation (44.1 kHz)
pub const SAMPLE_RATE: f32 = 44100.0;

/// Standard tuning reference frequency (A4 = 440 Hz)
pub const A4_FREQUENCY: f32 = 440.0;

/// Convert MIDI note number to frequency in Hz
///
/// # Arguments
/// * `midi_note` - MIDI note number (0-127, where 69 = A4 = 440Hz)
///
/// # Returns
/// Frequency in Hz
pub fn midi_to_frequency(midi_note: u8) -> f32 {
    A4_FREQUENCY * 2.0_f32.powf((midi_note as f32 - 69.0) / 12.0)
}

/// Convert frequency to MIDI note number
///
/// # Arguments
/// * `frequency` - Frequency in Hz
///
/// # Returns
/// MIDI note number (rounded to nearest integer)
pub fn frequency_to_midi(frequency: f32) -> u8 {
    (69.0 + 12.0 * (frequency / A4_FREQUENCY).log2()).round() as u8
}

/// Convert beats per minute to samples per beat
///
/// # Arguments
/// * `bpm` - Beats per minute
///
/// # Returns
/// Number of samples per beat at current sample rate
pub fn bpm_to_samples_per_beat(bpm: f32) -> usize {
    (SAMPLE_RATE * 60.0 / bpm) as usize
}

/// Normalize a value to the range [0.0, 1.0]
///
/// # Arguments
/// * `value` - Input value
/// * `min` - Minimum possible value
/// * `max` - Maximum possible value
///
/// # Returns
/// Normalized value between 0.0 and 1.0
pub fn normalize(value: f32, min: f32, max: f32) -> f32 {
    (value - min) / (max - min)
}

/// Linear interpolation between two values
///
/// # Arguments
/// * `a` - Start value
/// * `b` - End value
/// * `t` - Interpolation factor (0.0 to 1.0)
///
/// # Returns
/// Interpolated value
pub fn lerp(a: f32, b: f32, t: f32) -> f32 {
    a + (b - a) * t
}

/// Apply a simple low-pass filter to smooth audio transitions
///
/// # Arguments
/// * `input` - Input sample
/// * `previous` - Previous filtered sample
/// * `alpha` - Filter coefficient (0.0 to 1.0, higher = less filtering)
///
/// # Returns
/// Filtered sample
pub fn low_pass_filter(input: f32, previous: f32, alpha: f32) -> f32 {
    alpha * input + (1.0 - alpha) * previous
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_midi_to_frequency() {
        // A4 should be 440 Hz
        assert_eq!(midi_to_frequency(69), 440.0);

        // C4 should be approximately 261.63 Hz
        let c4_freq = midi_to_frequency(60);
        assert!((c4_freq - 261.63).abs() < 0.1);
    }

    #[test]
    fn test_frequency_to_midi() {
        assert_eq!(frequency_to_midi(440.0), 69);
        assert_eq!(frequency_to_midi(261.63), 60);
    }

    #[test]
    fn test_bpm_conversion() {
        let samples_per_beat = bpm_to_samples_per_beat(120.0);
        // At 120 BPM, each beat should be 0.5 seconds
        // At 44100 Hz sample rate, that's 22050 samples
        assert_eq!(samples_per_beat, 22050);
    }

    #[test]
    fn test_normalize() {
        assert_eq!(normalize(5.0, 0.0, 10.0), 0.5);
        assert_eq!(normalize(0.0, 0.0, 10.0), 0.0);
        assert_eq!(normalize(10.0, 0.0, 10.0), 1.0);
    }

    #[test]
    fn test_lerp() {
        assert_eq!(lerp(0.0, 10.0, 0.5), 5.0);
        assert_eq!(lerp(0.0, 10.0, 0.0), 0.0);
        assert_eq!(lerp(0.0, 10.0, 1.0), 10.0);
    }
}
