//! Generate electronic music without AI
//!
//! A command-line tool for generating electronic music without the use of AI.

use clap::{Parser, Subcommand};
use musicgen::audio::AudioExporter;
use musicgen::composition::Composition;
use musicgen::config::CompositionConfig;
use std::path::PathBuf;

#[derive(Parser)]
#[command(name = "musicgen")]
#[command(about = "Generate electronic music without AI")]
#[command(version = "0.1.0")]
struct Cli {
    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand)]
enum Commands {
    /// Generate composition from JSON configuration file
    Json {
        /// Path to JSON configuration file
        config: PathBuf,

        /// Override output filename
        #[arg(short, long)]
        output: Option<String>,

        /// Print example JSON configuration
        #[arg(long)]
        example: bool,
    },

    /// Show information about available options
    Info {
        /// Show available scales
        #[arg(long)]
        scales: bool,

        /// Show available instruments
        #[arg(long)]
        instruments: bool,

        /// Show all information
        #[arg(short, long)]
        all: bool,
    },
}

fn main() {
    let cli = Cli::parse();

    let result = match cli.command {
        Commands::Json {
            config,
            output,
            example,
        } => handle_json_command(config, output, example),

        Commands::Info {
            scales,
            instruments,
            all,
        } => show_info(scales, instruments, all),
    };

    if let Err(e) = result {
        eprintln!("Error: {}", e);
        std::process::exit(1);
    }
}

fn show_info(scales: bool, instruments: bool, all: bool) -> Result<(), Box<dyn std::error::Error>> {
    if all || scales {
        println!("Available Scales:");
        println!("  major       - Major scale (Ionian mode)");
        println!("  minor       - Natural minor scale (Aeolian mode)");
        println!("  dorian      - Dorian mode");
        println!("  phrygian    - Phrygian mode");
        println!("  lydian      - Lydian mode");
        println!("  mixolydian  - Mixolydian mode");
        println!("  aeolian     - Aeolian mode (same as minor)");
        println!("  locrian     - Locrian mode");
        println!("  pentatonic  - Pentatonic scale");
        println!("  blues       - Blues scale");
        println!("  chromatic   - Chromatic scale (all 12 notes)");
        println!();
    }

    if all || instruments {
        println!("Available Instruments:");
        println!("  sine        - Pure sine wave (smooth, warm)");
        println!("  square      - Square wave (classic electronic, punchy)");
        println!("  sawtooth    - Sawtooth wave (bright, buzzy)");
        println!("  triangle    - Triangle wave (mellow, soft)");
        println!("  noise       - White noise (for percussion, textures)");
        println!();
    }

    if all {
        println!("MIDI Note Numbers:");
        println!("  C4 = 60, C#4 = 61, D4 = 62, D#4 = 63, E4 = 64, F4 = 65");
        println!("  F#4 = 66, G4 = 67, G#4 = 68, A4 = 69, A#4 = 70, B4 = 71");
        println!("  (Add/subtract 12 for different octaves)");
        println!();

        println!("Pattern Types:");
        println!("  custom      - Specify exact notes with timing");
        println!("  chord       - Play chord progressions");
        println!("  arpeggio    - Arpeggiate chords");
        println!("  sequence    - Step sequencer patterns");
        println!();

        println!("Effect Types:");
        println!("  lowpass     - Low-pass filter (removes highs)");
        println!("  highpass    - High-pass filter (removes lows)");
        println!("  delay       - Echo effect");
        println!("  reverb      - Spatial reverb effect");
        println!("  chorus      - Modulation/thickening effect");
        println!("  distortion  - Harmonic distortion");
    }

    Ok(())
}

fn handle_json_command(
    config_path: PathBuf,
    output_override: Option<String>,
    show_example: bool,
) -> Result<(), Box<dyn std::error::Error>> {
    if show_example {
        let example = CompositionConfig::example();
        println!("{}", example.to_json_pretty()?);
        println!("\n# Save this to a .json file and run:");
        println!("# cargo run --bin musicgen json your_config.json");
        return Ok(());
    }

    println!("Loading configuration from: {}", config_path.display());

    let config = CompositionConfig::from_file(&config_path)?;

    println!("Composition: {}", config.metadata.title);
    println!("Key: {:?}", config.composition.key);
    println!("Scale: {}", config.composition.scale);
    println!("Tempo: {} BPM", config.composition.tempo);
    println!("Measures: {}", config.composition.measures);
    println!("Tracks: {}", config.tracks.len());

    // Use composition system
    let mut composition = Composition::from_config(config.clone())?;

    println!("\nGenerating composition...");
    composition.generate()?;

    let stats = composition.get_stats();
    println!(
        "Generated {} tracks with {} total notes",
        stats.track_count, stats.total_notes
    );

    // Export based on config
    let exporter = AudioExporter::default();
    let filename = output_override.unwrap_or(config.export.filename.clone());
    let filename_with_ext = if !filename.ends_with(".wav") {
        format!("{}.wav", filename)
    } else {
        filename
    };

    println!("\nExporting to: output/{}", filename_with_ext);

    // Convert to core Composition format for export
    let export_composition = composition.to_composition();

    if config.export.stereo {
        exporter.export_stereo_wav(
            &export_composition,
            &filename_with_ext,
            config.export.max_duration,
        )?;
    } else {
        exporter.export_wav(
            &export_composition,
            &filename_with_ext,
            config.export.max_duration,
        )?;
    }

    // Generate variations if requested
    if let Some(ref variations) = config.export.variations {
        println!("\nGenerating {} variations...", variations.count);
        let base_name = filename_with_ext.trim_end_matches(".wav");
        exporter.export_variations(
            &export_composition,
            base_name,
            variations.count,
            config.export.max_duration,
        )?;
    }

    println!("\nExport complete!");
    Ok(())
}
