//! Audio synthesis module for generating various waveforms and tones
//!
//! This module provides oscillators for generating basic waveforms like sine, square,
//! sawtooth, and triangle waves, as well as more complex synthesis techniques.

use crate::{PI, SAMPLE_RATE};
use std::f32::consts::TAU;

/// Different types of waveforms that can be generated
#[derive(Debug, Clone, Copy, PartialEq)]
pub enum Waveform {
    Sine,
    Square,
    Sawtooth,
    Triangle,
    Noise,
}

/// A basic oscillator that generates waveforms at a given frequency
#[derive(Debug, Clone)]
pub struct Oscillator {
    pub waveform: Waveform,
    pub frequency: f32,
    pub amplitude: f32,
    pub phase: f32,
    pub phase_increment: f32,
}

impl Oscillator {
    /// Create a new oscillator
    ///
    /// # Arguments
    /// * `waveform` - Type of waveform to generate
    /// * `frequency` - Frequency in Hz
    /// * `amplitude` - Amplitude (0.0 to 1.0)
    ///
    /// # Returns
    /// New oscillator instance
    pub fn new(waveform: Waveform, frequency: f32, amplitude: f32) -> Self {
        let phase_increment = frequency * TAU / SAMPLE_RATE;

        Self {
            waveform,
            frequency,
            amplitude,
            phase: 0.0,
            phase_increment,
        }
    }

    /// Set the frequency and update the phase increment
    pub fn set_frequency(&mut self, frequency: f32) {
        self.frequency = frequency;
        self.phase_increment = frequency * TAU / SAMPLE_RATE;
    }

    /// Set the amplitude
    pub fn set_amplitude(&mut self, amplitude: f32) {
        self.amplitude = amplitude.clamp(0.0, 1.0);
    }

    /// Generate the next sample
    pub fn next_sample(&mut self) -> f32 {
        let sample = match self.waveform {
            Waveform::Sine => self.sine_wave(),
            Waveform::Square => self.square_wave(),
            Waveform::Sawtooth => self.sawtooth_wave(),
            Waveform::Triangle => self.triangle_wave(),
            Waveform::Noise => self.noise_wave(),
        };

        // Advance phase
        self.phase += self.phase_increment;
        if self.phase >= TAU {
            self.phase -= TAU;
        }

        sample * self.amplitude
    }

    /// Generate multiple samples into a buffer
    pub fn fill_buffer(&mut self, buffer: &mut [f32]) {
        for sample in buffer.iter_mut() {
            *sample = self.next_sample();
        }
    }

    /// Reset the oscillator phase
    pub fn reset_phase(&mut self) {
        self.phase = 0.0;
    }

    fn sine_wave(&self) -> f32 {
        self.phase.sin()
    }

    fn square_wave(&self) -> f32 {
        if self.phase < PI { 1.0 } else { -1.0 }
    }

    fn sawtooth_wave(&self) -> f32 {
        (self.phase / PI) - 1.0
    }

    fn triangle_wave(&self) -> f32 {
        if self.phase < PI {
            (2.0 / PI) * self.phase - 1.0
        } else {
            3.0 - (2.0 / PI) * self.phase
        }
    }

    fn noise_wave(&self) -> f32 {
        use rand::Rng;
        let mut rng = rand::thread_rng();
        rng.gen_range(-1.0..1.0)
    }
}

/// An envelope generator for controlling amplitude over time
#[derive(Debug, Clone)]
pub struct EnvelopeGenerator {
    pub attack_time: f32,   // seconds
    pub decay_time: f32,    // seconds
    pub sustain_level: f32, // 0.0 to 1.0
    pub release_time: f32,  // seconds

    state: EnvelopeState,
    current_level: f32,
    time_in_state: f32,
}

#[derive(Debug, Clone, PartialEq)]
enum EnvelopeState {
    Idle,
    Attack,
    Decay,
    Sustain,
    Release,
}

impl EnvelopeGenerator {
    /// Create a new ADSR envelope generator
    ///
    /// # Arguments
    /// * `attack` - Attack time in seconds
    /// * `decay` - Decay time in seconds
    /// * `sustain` - Sustain level (0.0 to 1.0)
    /// * `release` - Release time in seconds
    pub fn new(attack: f32, decay: f32, sustain: f32, release: f32) -> Self {
        Self {
            attack_time: attack,
            decay_time: decay,
            sustain_level: sustain.clamp(0.0, 1.0),
            release_time: release,
            state: EnvelopeState::Idle,
            current_level: 0.0,
            time_in_state: 0.0,
        }
    }

    /// Trigger the envelope (start attack phase)
    pub fn trigger(&mut self) {
        self.state = EnvelopeState::Attack;
        self.time_in_state = 0.0;
    }

    /// Release the envelope (start release phase)
    pub fn release(&mut self) {
        if self.state != EnvelopeState::Idle && self.state != EnvelopeState::Release {
            self.state = EnvelopeState::Release;
            self.time_in_state = 0.0;
        }
    }

    /// Get the next envelope value
    pub fn next_sample(&mut self) -> f32 {
        let dt = 1.0 / SAMPLE_RATE;

        match self.state {
            EnvelopeState::Idle => {
                self.current_level = 0.0;
            }
            EnvelopeState::Attack => {
                if self.attack_time > 0.0 {
                    self.current_level = self.time_in_state / self.attack_time;
                    if self.current_level >= 1.0 {
                        self.current_level = 1.0;
                        self.state = EnvelopeState::Decay;
                        self.time_in_state = 0.0;
                    }
                } else {
                    self.current_level = 1.0;
                    self.state = EnvelopeState::Decay;
                    self.time_in_state = 0.0;
                }
            }
            EnvelopeState::Decay => {
                if self.decay_time > 0.0 {
                    let decay_progress = self.time_in_state / self.decay_time;
                    self.current_level = 1.0 - decay_progress * (1.0 - self.sustain_level);
                    if decay_progress >= 1.0 {
                        self.current_level = self.sustain_level;
                        self.state = EnvelopeState::Sustain;
                        self.time_in_state = 0.0;
                    }
                } else {
                    self.current_level = self.sustain_level;
                    self.state = EnvelopeState::Sustain;
                    self.time_in_state = 0.0;
                }
            }
            EnvelopeState::Sustain => {
                self.current_level = self.sustain_level;
            }
            EnvelopeState::Release => {
                if self.release_time > 0.0 {
                    let release_start_level = if self.time_in_state == 0.0 {
                        self.current_level
                    } else {
                        self.sustain_level
                    };

                    let release_progress = self.time_in_state / self.release_time;
                    self.current_level = release_start_level * (1.0 - release_progress);

                    if release_progress >= 1.0 {
                        self.current_level = 0.0;
                        self.state = EnvelopeState::Idle;
                        self.time_in_state = 0.0;
                    }
                } else {
                    self.current_level = 0.0;
                    self.state = EnvelopeState::Idle;
                    self.time_in_state = 0.0;
                }
            }
        }

        self.time_in_state += dt;
        self.current_level.clamp(0.0, 1.0)
    }

    /// Check if the envelope is finished (in idle state)
    pub fn is_finished(&self) -> bool {
        self.state == EnvelopeState::Idle
    }
}

/// A track combines an oscillator with an envelope for musical notes
#[derive(Debug, Clone)]
pub struct Track {
    pub oscillator: Oscillator,
    pub envelope: EnvelopeGenerator,
    pub is_active: bool,
}

impl Track {
    /// Create a new track
    pub fn new(waveform: Waveform, frequency: f32, amplitude: f32) -> Self {
        Self {
            oscillator: Oscillator::new(waveform, frequency, amplitude),
            envelope: EnvelopeGenerator::new(0.01, 0.1, 0.7, 0.3), // Default ADSR
            is_active: false,
        }
    }

    /// Trigger the track (start playing a note)
    pub fn trigger(&mut self, frequency: f32) {
        self.oscillator.set_frequency(frequency);
        self.oscillator.reset_phase();
        self.envelope.trigger();
        self.is_active = true;
    }

    /// Release the track (stop playing)
    pub fn release(&mut self) {
        self.envelope.release();
    }

    /// Generate the next sample
    pub fn next_sample(&mut self) -> f32 {
        if !self.is_active {
            return 0.0;
        }

        let osc_sample = self.oscillator.next_sample();
        let env_level = self.envelope.next_sample();

        if self.envelope.is_finished() {
            self.is_active = false;
        }

        osc_sample * env_level
    }

    /// Check if the track is currently active
    pub fn is_active(&self) -> bool {
        self.is_active && !self.envelope.is_finished()
    }
}

/// A polyphonic synthesizer that manages multiple tracks
#[derive(Debug)]
pub struct PolySynth {
    pub tracks: Vec<Track>,
    pub max_tracks: usize,
}

impl PolySynth {
    /// Create a new polyphonic synthesizer
    ///
    /// # Arguments
    /// * `max_tracks` - Maximum number of simultaneous tracks
    /// * `waveform` - Default waveform for all tracks
    pub fn new(max_tracks: usize, waveform: Waveform) -> Self {
        let mut tracks = Vec::with_capacity(max_tracks);
        for _ in 0..max_tracks {
            tracks.push(Track::new(waveform, 440.0, 0.5));
        }

        Self { tracks, max_tracks }
    }

    /// Play a note
    ///
    /// # Arguments
    /// * `frequency` - Frequency of the note in Hz
    /// Play a note on an available track
    pub fn play_note(&mut self, frequency: f32) {
        // Find an inactive track or steal the oldest one
        if let Some(track) = self.tracks.iter_mut().find(|t| !t.is_active()) {
            track.trigger(frequency);
        } else if let Some(track) = self.tracks.first_mut() {
            // Track stealing - use the first track
            track.trigger(frequency);
        }
    }

    /// Stop a note
    ///
    /// # Arguments
    /// * `frequency` - Frequency of the note to stop
    /// Stop a note with the given frequency
    pub fn stop_note(&mut self, frequency: f32) {
        const FREQ_TOLERANCE: f32 = 1.0; // Hz tolerance for frequency matching

        for track in &mut self.tracks {
            if track.is_active() && (track.oscillator.frequency - frequency).abs() < FREQ_TOLERANCE
            {
                track.release();
            }
        }
    }

    /// Generate the next sample (sum of all active tracks)
    pub fn next_sample(&mut self) -> f32 {
        let mut sample = 0.0;
        for track in &mut self.tracks {
            sample += track.next_sample();
        }

        // Normalize by number of tracks to prevent clipping
        sample / self.max_tracks as f32
    }

    /// Fill a buffer with samples
    pub fn fill_buffer(&mut self, buffer: &mut [f32]) {
        for sample in buffer.iter_mut() {
            *sample = self.next_sample();
        }
    }

    /// Get the number of currently active tracks
    pub fn active_track_count(&self) -> usize {
        self.tracks.iter().filter(|t| t.is_active()).count()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_oscillator_creation() {
        let osc = Oscillator::new(Waveform::Sine, 440.0, 0.5);
        assert_eq!(osc.frequency, 440.0);
        assert_eq!(osc.amplitude, 0.5);
        assert_eq!(osc.waveform, Waveform::Sine);
    }

    #[test]
    fn test_oscillator_sample_generation() {
        let mut osc = Oscillator::new(Waveform::Sine, 440.0, 1.0);
        let sample = osc.next_sample();
        assert!(sample >= -1.0 && sample <= 1.0);
    }

    #[test]
    fn test_envelope_generator() {
        let mut env = EnvelopeGenerator::new(0.1, 0.1, 0.5, 0.1);
        env.trigger();

        // Should start at 0 and increase during attack
        let initial = env.next_sample();
        assert!(initial >= 0.0);

        // Generate some samples to move through attack phase
        for _ in 0..1000 {
            env.next_sample();
        }

        env.release();
        let release_sample = env.next_sample();
        assert!(release_sample >= 0.0);
    }

    #[test]
    fn test_track_triggering() {
        let mut track = Track::new(Waveform::Sine, 440.0, 0.5);
        assert!(!track.is_active());

        track.trigger(440.0);
        assert!(track.is_active());

        track.release();
        // Track should still be active until envelope finishes
        assert!(track.is_active());
    }

    #[test]
    fn test_polysynth() {
        let mut synth = PolySynth::new(4, Waveform::Sine);
        assert_eq!(synth.active_track_count(), 0);

        synth.play_note(440.0);
        assert_eq!(synth.active_track_count(), 1);

        synth.play_note(880.0);
        assert_eq!(synth.active_track_count(), 2);

        // Allow envelope to ramp up by getting several samples
        let mut non_zero_found = false;
        for _ in 0..100 {
            let sample = synth.next_sample();
            if sample.abs() > 0.0 {
                non_zero_found = true;
                break;
            }
        }
        assert!(
            non_zero_found,
            "Expected to find non-zero samples from active tracks"
        );
    }
}
